/**
 * MCS Media Computer Software
 * Copyright 2018 by Wilfried Klaas
 * Project: SPSEmulator
 * File: Not.java
 * EMail: W.Klaas@gmx.de
 * Created: 25.11.2018 wklaa_000
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>
 */
package de.mcs.tools.sps.mnemonic;

import java.util.EnumSet;
import java.util.Set;

import org.apache.commons.lang3.StringUtils;

import de.mcs.tools.sps.exceptions.IllegalArgument;
import de.mcs.tools.sps.exceptions.SyntaxError;

/**
 * LDI #x: Load directly, setting register A to a static value. x in range
 * 0..15;
 * 
 * @author wklaa_000
 *
 */
public class LDA extends AbstractMnemonic implements Mnemonic {

  byte value;

  enum INPUTSOURCE {
    DIRECT, DIGITAL, ANALOG, RC
  };

  INPUTSOURCE inputSource;

  public LDA(String line) throws SyntaxError {
    super(line);
    value = 0;
    inputSource = INPUTSOURCE.DIRECT;
  }

  @Override
  public void checkArgument() throws SyntaxError {
    if (StringUtils.isEmpty(getArgument())) {
      throw new SyntaxError(getLineNumber(),
          String.format("missing argument for %s.", this.getClass().getSimpleName()));
    }
    if (!getArgument().startsWith("#")) {
      String inputString = getArgument().toUpperCase();
      if (inputString.startsWith("DIN")) {
        inputSource = INPUTSOURCE.DIGITAL;
        if ("DIN".equalsIgnoreCase(inputString)) {
          value = 0;
        } else {
          int myValue = Integer.parseInt(inputString.substring(3));
          if ((myValue < 1) || (myValue > 4)) {
            throw new IllegalArgument(getLineNumber(), String.format("argument %s is not in range DIN1..4 for %s.",
                getArgument(), this.getClass().getSimpleName()));
          }
          value = (byte) (myValue & 0xFF);
        }
      } else if (inputString.startsWith("ADC")) {
        inputSource = INPUTSOURCE.ANALOG;
        int myValue = 0;
        try {
          myValue = Integer.parseInt(inputString.substring(3));
        } catch (NumberFormatException e) {
        }
        if ((myValue < 1) || (myValue > 2)) {
          throw new IllegalArgument(getLineNumber(), String.format("argument %s is not in range ADC1..4 for %s.",
              getArgument(), this.getClass().getSimpleName()));
        }
        value = (byte) ((myValue - 1) & 0xFF);
      } else if (inputString.startsWith("RC")) {
        inputSource = INPUTSOURCE.RC;
        int myValue = 0;
        try {
          myValue = Integer.parseInt(inputString.substring(2));
        } catch (NumberFormatException e) {
        }
        if ((myValue < 1) || (myValue > 2)) {
          throw new IllegalArgument(getLineNumber(), String.format("argument %s is not in range RC1..4 for %s.",
              getArgument(), this.getClass().getSimpleName()));
        }
        value = (byte) ((myValue - 1) & 0xFF);
      } else {
        throw new IllegalArgument(getLineNumber(),
            String.format("argument %s sould start with \"#\".", getArgument(), this.getClass().getSimpleName()));
      }
    } else {
      String argument = getArgument();
      argument = argument.substring(1);
      int myValue = getArgumentAsNumber(argument);
      if ((myValue < 0) || (myValue > 15)) {
        throw new IllegalArgument(getLineNumber(),
            String.format("argument %s is not in range 0..15 for %s.", getArgument(), this.getClass().getSimpleName()));
      }
      value = (byte) (myValue & 0xFF);
    }
  }

  @Override
  public int getByte() {
    switch (inputSource) {
    case DIRECT:
      return 0x40 + value;
    case DIGITAL:
      return 0x64 + value;
    case ANALOG:
      return 0x69 + value;
    case RC:
      return 0x6b + value;
    default:
      return 0x40 + value;
    }
  }

  @Override
  public boolean isUsingPage() {
    return false;
  }

  @Override
  public boolean hasArgument() {
    return true;
  }

  @Override
  public Set<HARDWARE> allowedHardware() {
    switch (inputSource) {
    case DIRECT:
    case DIGITAL:
    case ANALOG:
      return super.allowedHardware();
    case RC:
      return EnumSet.of(HARDWARE.ARDUINOSPS, HARDWARE.TINYSPS);
    default:
      return super.allowedHardware();
    }
  }

}
