package de.mcs.tools.sps.emulator.emulator.holtek;

import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertThrows;

import java.util.ArrayList;
import java.util.List;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import de.mcs.tools.sps.emulator.EmulatorInternals;
import de.mcs.tools.sps.emulator.EmulatorOutput;
import de.mcs.tools.sps.emulator.SPSCommand;
import de.mcs.tools.sps.emulator.SPSCommandData;
import de.mcs.tools.sps.emulator.SPSEmulator;
import de.mcs.tools.sps.emulator.TOutputCallback;
import de.mcs.tools.sps.emulator.emulator.holtek.HoltekEmulator;
import de.mcs.tools.sps.emulator.emulator.holtek.HoltekEmulatorOutput;
import de.mcs.tools.sps.emulator.exceptions.WrongProgramSizeException;

class TestHoltekEmulator implements TOutputCallback {

  private SPSEmulator holtekEmulator;

  @BeforeEach
  void setUp() throws Exception {
    holtekEmulator = new HoltekEmulator();
    holtekEmulator.addCallback(this);
  }

  @Test
  void testOutputAllCommands() {
    System.out.println("command list holtect");
    List<SPSCommand> commands = holtekEmulator.getCommands();
    commands.forEach(c -> {
      List<SPSCommandData> commandDatas = c.getCommandDatas();
      System.out.printf("Command %s(%s):", c.getMnemonic(), c.getName());
      commandDatas.forEach(cd -> {
        System.out.printf("%s(%s) ,", cd.getMnemonic(), cd.getName());
      });
      System.out.println();
    });
  }

  @Test
  void testProgramSize() throws WrongProgramSizeException {
    assertEquals(128, holtekEmulator.getProgramSize());
    byte[] prgMemory = new byte[128];
    holtekEmulator.loadProgram(prgMemory);
  }

  @Test
  void testWrongProgramSize() throws WrongProgramSizeException {
    byte[] prgMemory = new byte[129];
    Throwable exception = assertThrows(WrongProgramSizeException.class, () -> holtekEmulator.loadProgram(prgMemory));
  }

  @Test
  void testCommandArray() {
    List<SPSCommand> commands = holtekEmulator.getCommands();
    assertEquals(15, commands.size());
    for (SPSCommand spsCommand : commands) {
      switch (spsCommand.getCommandByte()) {
      case 0x00:
        assertEquals(1, spsCommand.getCommandDatas().size());
        break;
      case 0x10:
        assertEquals(16, spsCommand.getCommandDatas().size());
        break;
      case 0x20:
        assertEquals(16, spsCommand.getCommandDatas().size());
        break;
      case 0x30:
        assertEquals(16, spsCommand.getCommandDatas().size());
        break;
      case 0x40:
        assertEquals(16, spsCommand.getCommandDatas().size());
        break;
      case 0x50:
        assertEquals(9, spsCommand.getCommandDatas().size());
        break;
      case 0x60:
        assertEquals(10, spsCommand.getCommandDatas().size());
        break;
      case 0x70:
        assertEquals(10, spsCommand.getCommandDatas().size());
        break;
      case (byte) 0x80:
        assertEquals(8, spsCommand.getCommandDatas().size());
        break;
      case (byte) 0x90:
        assertEquals(16, spsCommand.getCommandDatas().size());
        break;
      case (byte) 0xA0:
        assertEquals(16, spsCommand.getCommandDatas().size());
        break;
      case (byte) 0xB0:
        assertEquals(16, spsCommand.getCommandDatas().size());
        break;
      case (byte) 0xC0:
        assertEquals(15, spsCommand.getCommandDatas().size());
        break;
      case (byte) 0xD0:
        assertEquals(16, spsCommand.getCommandDatas().size());
        break;
      case (byte) 0xE0:
        assertEquals(1, spsCommand.getCommandDatas().size());
        break;
      default:
        fail(String.format("unknow command %s", spsCommand.getName()));
        break;
      }
    }
  }

  @Test
  void testNOP() throws WrongProgramSizeException {
    outputList.clear();
    internalList.clear();
    byte[] prg = new byte[] { 0x00, 0x00, 0x00 };
    holtekEmulator.loadProgram(prg);
    holtekEmulator.startProgram(true);
    holtekEmulator.nextStep();
    holtekEmulator.nextStep();
    holtekEmulator.nextStep();

    assertEquals(4, outputList.size());
    assertEquals(4, internalList.size());
  }

  List<EmulatorOutput> outputList = new ArrayList();
  List<EmulatorInternals> internalList = new ArrayList<>();

  @Override
  public void onOutput(EmulatorOutput output) {
    outputList.add(output);
    internalList.add(holtekEmulator.getEmulatorInternals());
  }

  @Test
  void testPort() throws WrongProgramSizeException {
    byte[] prg = new byte[] { 0x10, 0x1F, 0x10 };
    holtekEmulator.loadProgram(prg);
    holtekEmulator.startProgram(true);
    holtekEmulator.nextStep();
    HoltekEmulatorOutput emulatorOutput = (HoltekEmulatorOutput) holtekEmulator.getEmulatorOutput();
    assertFalse(emulatorOutput.d0);
    assertFalse(emulatorOutput.d1);
    assertFalse(emulatorOutput.d2);
    assertFalse(emulatorOutput.d3);
    holtekEmulator.nextStep();
    assertTrue(emulatorOutput.d0);
    assertTrue(emulatorOutput.d1);
    assertTrue(emulatorOutput.d2);
    assertTrue(emulatorOutput.d3);
    holtekEmulator.nextStep();
    assertFalse(emulatorOutput.d0);
    assertFalse(emulatorOutput.d1);
    assertFalse(emulatorOutput.d2);
    assertFalse(emulatorOutput.d3);
  }

  @Test
  void testDelay() throws WrongProgramSizeException {
    byte[] prg = new byte[] { 0x29, 0x2C };
    holtekEmulator.loadProgram(prg);
    holtekEmulator.startProgram(true);
    long start = System.currentTimeMillis();
    holtekEmulator.nextStep();
    long delta = System.currentTimeMillis() - start;
    assertTrue(delta > 900 && delta < 1100);
    start = System.currentTimeMillis();
    holtekEmulator.nextStep();
    delta = System.currentTimeMillis() - start;
    assertTrue(delta > 9000 && delta < 11000);
  }

  @Test
  void testJumpDown() throws WrongProgramSizeException {
    byte[] prg = new byte[] { 0x00, 0x00, 0x32 };
    holtekEmulator.loadProgram(prg);
    holtekEmulator.startProgram(true);
    holtekEmulator.nextStep();
    holtekEmulator.nextStep();
    int addr = holtekEmulator.getEmulatorInternals().getAddress();
    holtekEmulator.nextStep();
    assertEquals(0x00, holtekEmulator.getEmulatorInternals().getAddress());

    prg = new byte[] { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x39 };
    holtekEmulator.loadProgram(prg);
    holtekEmulator.startProgram(true);
    holtekEmulator.nextStep();
    holtekEmulator.nextStep();
    addr = holtekEmulator.getEmulatorInternals().getAddress();
    holtekEmulator.nextStep();
    assertEquals(0x03, holtekEmulator.getEmulatorInternals().getAddress());
  }

  @Test
  void testAEquals() throws WrongProgramSizeException {
    byte[] prg = new byte[16];
    for (int i = 0; i < prg.length; i++) {
      prg[i] = (byte) (0x40 + i);
    }
    ;
    holtekEmulator.loadProgram(prg);
    holtekEmulator.startProgram(true);
    for (int i = 0; i < prg.length; i++) {
      holtekEmulator.nextStep();
      assertEquals(i, holtekEmulator.getEmulatorInternals().getRegister("A"));
    }
  }

  @Test
  void testEqualsA() throws WrongProgramSizeException {
    byte[] prg = new byte[] { 0x48, 0x51, 0x52, 0x53, 0x40, 0x54, 0x4F, 0x54, 0x55 };
    holtekEmulator.loadProgram(prg);
    holtekEmulator.startProgram(true);
    holtekEmulator.nextStep();
    holtekEmulator.nextStep();
    holtekEmulator.nextStep();
    holtekEmulator.nextStep();
    assertEquals(0x08, holtekEmulator.getEmulatorInternals().getRegister("A"));
    assertEquals(0x08, holtekEmulator.getEmulatorInternals().getRegister("B"));
    assertEquals(0x08, holtekEmulator.getEmulatorInternals().getRegister("C"));
    assertEquals(0x08, holtekEmulator.getEmulatorInternals().getRegister("D"));

    holtekEmulator.nextStep();
    holtekEmulator.nextStep();
    assertEquals(0x00, ((HoltekEmulatorOutput) holtekEmulator.getEmulatorOutput()).getOutput());

    holtekEmulator.nextStep();
    holtekEmulator.nextStep();
    assertEquals(0x0F, ((HoltekEmulatorOutput) holtekEmulator.getEmulatorOutput()).getOutput());
    assertTrue(((HoltekEmulatorOutput) holtekEmulator.getEmulatorOutput()).d0);
    assertTrue(((HoltekEmulatorOutput) holtekEmulator.getEmulatorOutput()).d1);
    assertTrue(((HoltekEmulatorOutput) holtekEmulator.getEmulatorOutput()).d2);
    assertTrue(((HoltekEmulatorOutput) holtekEmulator.getEmulatorOutput()).d3);

    holtekEmulator.stop();

    prg = new byte[] { 0x41, 0x55, 0x56, 0x57, 0x58 };

    holtekEmulator.loadProgram(prg);
    holtekEmulator.startProgram(true);
    holtekEmulator.nextStep();
    holtekEmulator.nextStep();
    assertEquals(0x01, ((HoltekEmulatorOutput) holtekEmulator.getEmulatorOutput()).getOutput());
    holtekEmulator.nextStep();
    assertEquals(0x03, ((HoltekEmulatorOutput) holtekEmulator.getEmulatorOutput()).getOutput());
    assertTrue(((HoltekEmulatorOutput) holtekEmulator.getEmulatorOutput()).d0);
    assertTrue(((HoltekEmulatorOutput) holtekEmulator.getEmulatorOutput()).d1);
    assertFalse(((HoltekEmulatorOutput) holtekEmulator.getEmulatorOutput()).d2);
    assertFalse(((HoltekEmulatorOutput) holtekEmulator.getEmulatorOutput()).d3);
    holtekEmulator.nextStep();
    assertEquals(0x07, ((HoltekEmulatorOutput) holtekEmulator.getEmulatorOutput()).getOutput());
    holtekEmulator.nextStep();
    assertEquals(0x0f, ((HoltekEmulatorOutput) holtekEmulator.getEmulatorOutput()).getOutput());

    holtekEmulator.stop();

    prg = new byte[] { 0x41, 0x59, 0x4F, 0x59 };

    holtekEmulator.loadProgram(prg);
    holtekEmulator.startProgram(true);
    holtekEmulator.nextStep();
    holtekEmulator.nextStep();
    assertEquals(0x01, ((HoltekEmulatorOutput) holtekEmulator.getEmulatorOutput()).getFeature("PWM"));
    holtekEmulator.nextStep();
    holtekEmulator.nextStep();
    assertEquals(0x0F, ((HoltekEmulatorOutput) holtekEmulator.getEmulatorOutput()).getFeature("PWM"));
  }

  @Test
  void testAInput() throws WrongProgramSizeException {
    byte[] prg = new byte[] { 0x4F, 0x51, 0x52, 0x53, 0x40, 0x61, 0x40, 0x62, 0x40, 0x63 };
    holtekEmulator.loadProgram(prg);
    holtekEmulator.startProgram(true);
    for (int i = 0; i < 5; i++) {
      holtekEmulator.nextStep();
    }
    assertEquals(0x00, holtekEmulator.getEmulatorInternals().getRegister("A"));
    holtekEmulator.nextStep();
    assertEquals(0x0f, holtekEmulator.getEmulatorInternals().getRegister("A"));
    holtekEmulator.nextStep();
    assertEquals(0x00, holtekEmulator.getEmulatorInternals().getRegister("A"));
    holtekEmulator.nextStep();
    assertEquals(0x0f, holtekEmulator.getEmulatorInternals().getRegister("A"));
    holtekEmulator.nextStep();
    assertEquals(0x00, holtekEmulator.getEmulatorInternals().getRegister("A"));
    holtekEmulator.nextStep();
    assertEquals(0x0f, holtekEmulator.getEmulatorInternals().getRegister("A"));

    holtekEmulator.stop();

    prg = new byte[] { 0x64, 0x64, 0x40, 0x65, 0x40, 0x66, 0x40, 0x67, 0x40, 0x68 };
    holtekEmulator.loadProgram(prg);
    holtekEmulator.startProgram(true);
    holtekEmulator.getEmulatorInput().setInput((byte) 0x00);
    holtekEmulator.nextStep();
    assertEquals(0x00, holtekEmulator.getEmulatorInternals().getRegister("A"));
    holtekEmulator.getEmulatorInput().setInput((byte) 0x0f);
    holtekEmulator.nextStep();
    assertEquals(0x0f, holtekEmulator.getEmulatorInternals().getRegister("A"));

    holtekEmulator.nextStep();
    assertEquals(0x00, holtekEmulator.getEmulatorInternals().getRegister("A"));
    holtekEmulator.nextStep();
    assertEquals(0x01, holtekEmulator.getEmulatorInternals().getRegister("A"));

    holtekEmulator.nextStep();
    assertEquals(0x00, holtekEmulator.getEmulatorInternals().getRegister("A"));
    holtekEmulator.nextStep();
    assertEquals(0x01, holtekEmulator.getEmulatorInternals().getRegister("A"));

    holtekEmulator.nextStep();
    assertEquals(0x00, holtekEmulator.getEmulatorInternals().getRegister("A"));
    holtekEmulator.nextStep();
    assertEquals(0x01, holtekEmulator.getEmulatorInternals().getRegister("A"));

    holtekEmulator.nextStep();
    assertEquals(0x00, holtekEmulator.getEmulatorInternals().getRegister("A"));
    holtekEmulator.nextStep();
    assertEquals(0x01, holtekEmulator.getEmulatorInternals().getRegister("A"));

    holtekEmulator.stop();

    prg = new byte[] { 0x40, 0x69, 0x40, 0x69, 0x40, 0x6A, 0x40, 0x6A };
    holtekEmulator.loadProgram(prg);
    holtekEmulator.startProgram(true);
    holtekEmulator.nextStep();
    assertEquals(0x00, holtekEmulator.getEmulatorInternals().getRegister("A"));
    holtekEmulator.getEmulatorInput().setFeature("AD1", (byte) 0x00);
    holtekEmulator.nextStep();
    assertEquals(0x00, holtekEmulator.getEmulatorInternals().getRegister("A"));
    holtekEmulator.nextStep();
    assertEquals(0x00, holtekEmulator.getEmulatorInternals().getRegister("A"));
    holtekEmulator.getEmulatorInput().setFeature("AD1", (byte) 0x0F);
    holtekEmulator.nextStep();
    assertEquals(0x0f, holtekEmulator.getEmulatorInternals().getRegister("A"));

    holtekEmulator.nextStep();
    assertEquals(0x00, holtekEmulator.getEmulatorInternals().getRegister("A"));
    holtekEmulator.getEmulatorInput().setFeature("AD2", (byte) 0x00);
    holtekEmulator.nextStep();
    assertEquals(0x00, holtekEmulator.getEmulatorInternals().getRegister("A"));
    holtekEmulator.nextStep();
    assertEquals(0x00, holtekEmulator.getEmulatorInternals().getRegister("A"));
    holtekEmulator.getEmulatorInput().setFeature("AD2", (byte) 0x0F);
    holtekEmulator.nextStep();
    assertEquals(0x0f, holtekEmulator.getEmulatorInternals().getRegister("A"));
  }
}
